/**
 * Copyright 2021 Huawei Technologies Co., Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <gtest/gtest.h>

#define protected public
#define private public
#include "graph/passes/infer_value_range_pass.h"
#include "graph/utils/tensor_utils.h"
#include "graph/utils/graph_utils.h"
#include "graph_builder_utils.h"

#include "inc/external/graph/operator_reg.h"
#include "inc/external/graph/operator.h"
#include "inc/external/graph/operator_factory.h"
#include "inc/graph/operator_factory_impl.h"
#include "inc/kernel.h"
#include "inc/kernel_factory.h"

using namespace std;
using namespace testing;
namespace ge {
class UtestGraphInferValueRangePass : public testing::Test {
 protected:
  void SetUp() {}
  void TearDown() {}
};

static NodePtr CreateNode(ComputeGraph &graph, const string &name, const string &type, int in_num, int out_num) {
  auto op_desc = std::make_shared<OpDesc>("AddN", "AddN");
  return graph.AddNode(op_desc);
}

TEST_F(UtestGraphInferValueRangePass, infer_pass_not_register) {
  auto graph = std::make_shared<ComputeGraph>("test_graph");
  GeTensorDesc ge_tensor_desc(GeShape({1, 1, 4, 192}), ge::FORMAT_NCHW, DT_FLOAT16);
  auto addn_op_desc = std::make_shared<OpDesc>("AddN", "AddN");
  addn_op_desc->AddInputDesc(ge_tensor_desc);
  addn_op_desc->AddOutputDesc(ge_tensor_desc);
  auto addn_op_node = graph->AddNode(addn_op_desc);

  InferValueRangePass infer_pass;
  EXPECT_EQ(infer_pass.Run(addn_op_node), SUCCESS);
}

auto ShapeValueInfer = [&](Operator &op) {
  auto op_desc = OpDescUtils::GetOpDescFromOperator(op);
  auto output_tensor_desc = op_desc->MutableOutputDesc(0);
  std::vector<std::pair<int64_t, int64_t>> in_shape_range;
  op_desc->MutableInputDesc(0)->GetShapeRange(in_shape_range);
  if (!in_shape_range.empty()) {
    output_tensor_desc->SetValueRange(in_shape_range);
  }
  return SUCCESS;
};

TEST_F(UtestGraphInferValueRangePass, infer_pass_when_call_1_not_infer) {
  INFER_VALUE_RANGE_CUSTOM_FUNC_REG(Shape, INPUT_IS_DYNAMIC, ShapeValueInfer);
  auto graph = std::make_shared<ComputeGraph>("test_graph");
  GeTensorDesc ge_tensor_desc(GeShape({1, 1, 4, 192}), ge::FORMAT_NCHW, DT_INT32);
  std::vector<std::pair<int64_t, int64_t>> shape_range = {make_pair(1, 1), make_pair(1, 1),
                                                          make_pair(4, 4),   make_pair(192, 192)};
  ge_tensor_desc.SetShapeRange(shape_range);
  GeTensorDesc output_tensor_desc(GeShape({4}), ge::FORMAT_NCHW, DT_INT32);
  auto op_desc = std::make_shared<OpDesc>("Shape", "Shape");
  op_desc->AddInputDesc(ge_tensor_desc);
  op_desc->AddOutputDesc(output_tensor_desc);
  auto op_node = graph->AddNode(op_desc);

  InferValueRangePass infer_pass;
  EXPECT_EQ(infer_pass.Run(op_node), SUCCESS);

  auto output_0_desc = op_node->GetOpDesc()->GetOutputDesc(0);
  std::vector<std::pair<int64_t, int64_t>> value_range;
  output_0_desc.GetValueRange(value_range);
  EXPECT_EQ(value_range.empty(), true);
}

TEST_F(UtestGraphInferValueRangePass, infer_pass_when_call_1_infer) {
  // sqrt -> shape -> Output
  INFER_VALUE_RANGE_CUSTOM_FUNC_REG(Shape, INPUT_IS_DYNAMIC, ShapeValueInfer);
  auto graph = std::make_shared<ComputeGraph>("test_graph");
  GeTensorDesc sqrt_tensor_desc(GeShape({-1, -1, 4, 192}), ge::FORMAT_NCHW, DT_INT32);
  std::vector<std::pair<int64_t, int64_t>> shape_range = {make_pair(1, 100), make_pair(1, 240),
                                                          make_pair(4, 4),   make_pair(192, 192)};
  sqrt_tensor_desc.SetShapeRange(shape_range);
  auto sqrt_op_desc = std::make_shared<OpDesc>("Sqrt", "Sqrt");
  sqrt_op_desc->AddInputDesc(sqrt_tensor_desc);
  sqrt_op_desc->AddOutputDesc(sqrt_tensor_desc);
  auto sqrt_node = graph->AddNode(sqrt_op_desc);

  GeTensorDesc shape_output_desc(GeShape({4}), ge::FORMAT_NCHW, DT_INT32);
  auto shape_op_desc = std::make_shared<OpDesc>("Shape", "Shape");
  shape_op_desc->AddInputDesc(sqrt_tensor_desc);
  shape_op_desc->AddOutputDesc(shape_output_desc);
  auto shape_node = graph->AddNode(shape_op_desc);

  GeTensorDesc Output_in_tensor_desc(GeShape({4}), ge::FORMAT_NCHW, ge::DT_INT32);
  auto Output_op_desc = std::make_shared<OpDesc>("Output", "Output");
  Output_op_desc->AddInputDesc(Output_in_tensor_desc);
  auto Output_node = graph->AddNode(Output_op_desc);

  ge::GraphUtils::AddEdge(sqrt_node->GetOutDataAnchor(0), shape_node->GetInDataAnchor(0));
  ge::GraphUtils::AddEdge(shape_node->GetOutDataAnchor(0), Output_node->GetInDataAnchor(0));
  EXPECT_EQ(graph->TopologicalSorting(), GRAPH_SUCCESS);


  InferValueRangePass infer_pass;
  auto ret = infer_pass.Run(shape_node);
  EXPECT_EQ(ret, SUCCESS);

  auto output_0_desc = shape_node->GetOpDesc()->GetOutputDesc(0);
  std::vector<std::pair<int64_t, int64_t>> value_range;
  output_0_desc.GetValueRange(value_range);
  EXPECT_EQ(value_range.size(), 4);
  std::vector<int64_t> target_value_range = {1, 100, 1, 240, 4, 4, 192, 192};
  std::vector<int64_t> output_value_range;
  for (auto pair : value_range) {
    output_value_range.push_back(pair.first);
    output_value_range.push_back(pair.second);
  }
  EXPECT_EQ(target_value_range, output_value_range);

  auto in_0_desc = Output_node->GetOpDesc()->GetInputDesc(0);
  value_range.clear();
  in_0_desc.GetValueRange(value_range);
  EXPECT_EQ(value_range.size(), 0);

  /*
  INFER_VALUE_RANGE_DEFAULT_REG(Output);
  ret = infer_pass.Run(Output_node);
  EXPECT_EQ(ret, FAILED);
  auto in_0_desc_after_infer = Output_node->GetOpDesc()->GetInputDesc(0);
  value_range.clear();
  in_0_desc_after_infer.GetValueRange(value_range);
  EXPECT_EQ(value_range.size(), 4);
  output_value_range.clear();
  for (auto pair : value_range) {
    output_value_range.push_back(pair.first);
    output_value_range.push_back(pair.second);
  }
  EXPECT_EQ(target_value_range, output_value_range);

  auto out_0_desc = Output_node->GetOpDesc()->GetOutputDesc(0);
  value_range.clear();
  out_0_desc.GetValueRange(value_range);
  EXPECT_EQ(value_range.size(), 0);
   */
}

class AddKernel : public Kernel {
 public:
  Status Compute(const ge::OpDescPtr op_desc_ptr, const std::vector<ge::ConstGeTensorPtr> &input,
                 std::vector<ge::GeTensorPtr> &v_output) override {
    vector<int64_t> data_vec;
    auto data_num = input[0]->GetTensorDesc().GetShape().GetShapeSize();
    auto x1_data = reinterpret_cast<const int64_t *>(input[0]->GetData().data());
    auto x2_data = reinterpret_cast<const int64_t *>(input[1]->GetData().data());
    for (size_t i = 0; i < data_num; i++) {
      auto x_index = *(x1_data + i);
      auto y_index = *(x2_data + i);
      data_vec.push_back(x_index + y_index);
    }
    GeTensorPtr const_tensor = std::make_shared<ge::GeTensor>(input[0]->GetTensorDesc(), (uint8_t *)data_vec.data(),
                                                              data_num * sizeof(int64_t));
    v_output.emplace_back(const_tensor);
    return SUCCESS;
  }
};
REGISTER_KERNEL(ADD, AddKernel);

TEST_F(UtestGraphInferValueRangePass, infer_pass_when_call_2_infer) {
  // shape --- add --- sqrt
  // constant /
  INFER_VALUE_RANGE_DEFAULT_REG(Add);
  auto graph = std::make_shared<ComputeGraph>("test_graph");

  vector<int64_t> dims_vec = {4};
  vector<int64_t> data_vec = {1, 1, 1, 1};
  GeTensorDesc const_tensor_desc(ge::GeShape(dims_vec), ge::FORMAT_NCHW, ge::DT_INT64);
  GeTensorPtr const_tensor =
    std::make_shared<ge::GeTensor>(const_tensor_desc, (uint8_t *)data_vec.data(), data_vec.size() * sizeof(int64_t));

  auto const_op_desc = std::make_shared<OpDesc>("Constant", "Constant");
  const_op_desc->AddOutputDesc(const_tensor_desc);
  EXPECT_EQ(OpDescUtils::SetWeights(const_op_desc, const_tensor), GRAPH_SUCCESS);
  auto const_node = graph->AddNode(const_op_desc);

  GeTensorDesc shape_tensor_desc(GeShape({4}), ge::FORMAT_NCHW, ge::DT_INT64);
  std::vector<std::pair<int64_t, int64_t>> value_range = {make_pair(1, 100), make_pair(1, 240),
                                                          make_pair(4, 4),   make_pair(192, 192)};
  shape_tensor_desc.SetValueRange(value_range);
  auto shape_op_desc = std::make_shared<OpDesc>("Shape", "Shape");
  shape_op_desc->AddOutputDesc(shape_tensor_desc);
  auto shape_node = graph->AddNode(shape_op_desc);

  GeTensorDesc add_tensor_desc(GeShape({4}), ge::FORMAT_NCHW, ge::DT_INT64);
  auto add_op_desc = std::make_shared<OpDesc>("Add", "Add");
  add_op_desc->AddInputDesc(shape_tensor_desc);
  add_op_desc->AddInputDesc(const_tensor_desc);
  add_op_desc->AddOutputDesc(add_tensor_desc);
  auto add_node = graph->AddNode(add_op_desc);

  ge::GraphUtils::AddEdge(shape_node->GetOutDataAnchor(0), add_node->GetInDataAnchor(0));
  ge::GraphUtils::AddEdge(const_node->GetOutDataAnchor(0), add_node->GetInDataAnchor(1));

  InferValueRangePass infer_pass;
  EXPECT_EQ(infer_pass.Run(add_node), SUCCESS);

  auto output_0_desc = add_node->GetOpDesc()->GetOutputDesc(0);
  std::vector<std::pair<int64_t, int64_t>> out_value_range;
  output_0_desc.GetValueRange(out_value_range);
  EXPECT_EQ(out_value_range.size(), 4);

  std::vector<int64_t> target_value_range = {2, 101, 2, 241, 5, 5, 193, 193};
  std::vector<int64_t> output_value_range;
  for (auto pair : out_value_range) {
    output_value_range.push_back(pair.first);
    output_value_range.push_back(pair.second);
  }
  EXPECT_EQ(target_value_range, output_value_range);
}

TEST_F(UtestGraphInferValueRangePass, test_value_range_infer_and_set_get) {
  using std::make_pair;
  std::function<ge::graphStatus(ge::Operator &)> ShapeValueInfer_ = [](ge::Operator &op) -> ge::graphStatus {
    auto op_desc = OpDescUtils::GetOpDescFromOperator(op);
    auto output_tensor_desc = op_desc->MutableOutputDesc(0);
    std::vector<std::pair<int64_t, int64_t>> in_shape_range;
    op_desc->MutableInputDesc(0)->GetShapeRange(in_shape_range);
    if (!in_shape_range.empty()) {
      output_tensor_desc->SetValueRange(in_shape_range);
    }
    return GRAPH_SUCCESS;
  };

  INFER_VALUE_RANGE_CUSTOM_FUNC_REG(Shape, INPUT_IS_DYNAMIC, ShapeValueInfer_);
  string op_type = "Shape";

  auto graph = std::make_shared<ComputeGraph>("test_graph");
  auto shape_op_desc = std::make_shared<OpDesc>("node_name", op_type);
  GeTensorDesc tensor_desc(GeShape({-1, -1, 4, 192}), ge::FORMAT_NCHW, DT_INT32);
  std::vector<std::pair<int64_t, int64_t>> shape_range = {make_pair(1, 100), make_pair(1, 240),
                                                          make_pair(4, 4),   make_pair(192, 192)};
  tensor_desc.SetShapeRange(shape_range);
  shape_op_desc->AddInputDesc(tensor_desc);
  GeTensorDesc out_tensor_desc(GeShape({4}), ge::FORMAT_NCHW, DT_INT32);
  shape_op_desc->AddOutputDesc(out_tensor_desc);
  auto shape_node = graph->AddNode(shape_op_desc);
  Operator op = OpDescUtils::CreateOperatorFromNode(shape_node);
  auto ret = shape_node->GetOpDesc()->CallInferValueRangeFunc(op);
  ASSERT_EQ(ret, GRAPH_SUCCESS);

  auto output_0_desc = shape_node->GetOpDesc()->GetOutputDesc(0);
  std::vector<std::pair<int64_t, int64_t>> value_range;
  output_0_desc.GetValueRange(value_range);
  EXPECT_EQ(value_range.size(), 4);

  std::vector<int64_t> target_value_range = {1, 100, 1, 240, 4, 4, 192, 192};
  std::vector<int64_t> output_value_range;
  for (auto pair : value_range) {
    output_value_range.push_back(pair.first);
    output_value_range.push_back(pair.second);
  }
  EXPECT_EQ(target_value_range, output_value_range);
}

}  // namespace ge
